#!/usr/bin/env python3

"""
Created on 17 Oct 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The access_key utility is used to manage an AWS access key. The key is made up of a key ID (AWS_ACCESS_KEY_ID)
and a secret key (AWS_SECRET_ACCESS_KEY). The JSON key document managed by this utility is encrypted.

A password must be specified when the key is created and is required when the key is accessed.

SYNOPSIS
access_key.py [{ -s | -d }] [-v]

EXAMPLES
./access_key.py -s

FILES
~/SCS/aws/access_key.json

DOCUMENT EXAMPLE
{"key-id": "123", "secret-key": "456"}

SEE ALSO
scs_analysis/aws_bucket

RESOURCES
https://stackoverflow.com/questions/42568262/how-to-encrypt-text-with-a-password-in-python
"""

import sys

from scs_analysis.cmd.cmd_access_key import CmdAccessKey

from scs_core.aws.client.access_key import AccessKey
from scs_core.data.json import JSONify
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    access_key = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdAccessKey()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('access_key', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.set:
            access_key = AccessKey.from_user()
            password = AccessKey.password_from_user()

            access_key.save(Host, encryption_key=password)

        elif cmd.delete:
            AccessKey.delete(Host)

        else:
            password = AccessKey.password_from_user()

            try:
                access_key = AccessKey.load(Host, encryption_key=password)
            except (KeyError, ValueError):
                logger.error("incorrect password")
                exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # end...

        if access_key:
            print(JSONify.dumps(access_key))

    except KeyboardInterrupt:
        print(file=sys.stderr)



"""
Created on 13 Jul 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdSocketReceiver(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-p PORT] [-v]", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--port", "-p", type="int", nargs=1, action="store", default=2000, dest="port",
                                 help="socket port (default 2000)")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def port(self):
        return self.__opts.port


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CmdSocketReceiver:{port:%d, verbose:%s}" % (self.port, self.verbose)



"""
Created on 24 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdCognitoUserIdentity(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-c CREDENTIALS] | -C | -R | -U } [-i INDENT] [-v]",
                                              version="%prog 1.0")

        # identity...
        self.__parser.add_option("--credentials", "-c", type="string", action="store", dest="credentials_name",
                                 help="the stored credentials to be presented")

        # operations...
        self.__parser.add_option("--Create", "-C", action="store_true", dest="create", default=False,
                                 help="create my identity")

        self.__parser.add_option("--Retrieve", "-R", action="store_true", dest="retrieve", default=False,
                                 help="retrieve my identity")

        self.__parser.add_option("--Update", "-U", action="store_true", dest="update", default=False,
                                 help="update my identity")

        # output...
        self.__parser.add_option("--indent", "-i", type="int", nargs=1, action="store", dest="indent",
                                 help="pretty-print the output with INDENT")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.create:
            count += 1

        if self.retrieve:
            count += 1

        if self.update:
            count += 1

        if count != 1:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def credentials_name(self):
        return self.__opts.credentials_name


    @property
    def create(self):
        return self.__opts.create


    @property
    def retrieve(self):
        return self.__opts.retrieve


    @property
    def update(self):
        return self.__opts.update


    @property
    def indent(self):
        return self.__opts.indent


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdCognitoUserIdentity:{credentials_name:%s, retrieve:%s, create:%s, update:%s, indent:%s, " \
               "verbose:%s}" % \
               (self.credentials_name, self.retrieve, self.create, self.update, self.indent,
                self.verbose)



"""
Created on 28 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse

from scs_core.data.recurring_period import RecurringMinutes
from scs_core.estate.baseline_conf import BaselineConf
from scs_core.gas.minimum import Minimum


# --------------------------------------------------------------------------------------------------------------------

class CmdBaseline(object):
    """unix command line handler"""

    __UPTAKE_CMDS = ('restart', 'reboot', 'shutdown')
    __DEFAULT_UPTAKE_CMD = 'shutdown'

    # --------------------------------------------------------------------------------------------------------------------

    def __init__(self):
        """
        Constructor
        """
        cmds = ' | '.join(self.__UPTAKE_CMDS)

        self.__parser = optparse.OptionParser(usage="%prog [-a] -c NAME -f { V | E } [{ -r | -u COMMAND }] "
                                                    "[-s START] [-e END] [-p AGGREGATION] [-m GAS MINIMUM] "
                                                    "[{ -o GAS | -x GAS }] [-v] DEVICE_TAG_1 .. DEVICE_TAG_N",
                                              version="%prog 1.0")

        # source...
        self.__parser.add_option("--aws", "-a", action="store_true", dest="aws", default=False,
                                 help="Use AWS S3 instead of local storage for configuration")

        # identity...
        self.__parser.add_option("--conf-name", "-c", type="string", nargs=1, action="store", dest="conf_name",
                                 help="the name of the baseline configuration")

        # function...
        self.__parser.add_option("--fields", "-f", type="string", nargs=1, action="store", dest="fields",
                                 help="baseline Val or Exg fields")

        self.__parser.add_option("--rehearse", "-r", action="store_true", dest="rehearse", default=False,
                                 help="show what actions should be performed")

        self.__parser.add_option("--uptake", "-u", type="string", nargs=1, action="store", dest="uptake",
                                 default=self.__DEFAULT_UPTAKE_CMD,
                                 help="{ %s }, default '%s'" % (cmds, self.__DEFAULT_UPTAKE_CMD))

        # fields...
        self.__parser.add_option("--start-hour", "-s", type="int", nargs=1, action="store", dest="start_hour",
                                 help="override the configuration's start hour")

        self.__parser.add_option("--end-hour", "-e", type="int", nargs=1, action="store", dest="end_hour",
                                 help="override the configuration's end hour")

        self.__parser.add_option("--aggregation-period", "-p", type="int", nargs=1, action="store",
                                 dest="aggregation_period", help="override the configuration's aggregation period")

        self.__parser.add_option("--minimum", "-m", type="string", nargs=2, action="store", dest="minimum",
                                 help="override configuration's minimum for GAS")

        # exclusions...
        self.__parser.add_option("--only-gas", "-o", type="string", nargs=1, action="store", dest="only_gas",
                                 help="baseline only GAS")

        self.__parser.add_option("--exclude-gas", "-x", type="string", nargs=1, action="store", dest="exclude_gas",
                                 help="exclude GAS")

        # output...
        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.conf_name is None or len(self.device_tags) < 1:
            return False

        if self.fields not in Minimum.FIELD_SELECTIONS:
            return False

        if self.uptake not in self.__UPTAKE_CMDS:
            return False

        try:
            _ = self.minimum_value
        except ValueError:
            return False

        if self.only_gas is not None and self.exclude_gas is not None:
            return False

        return True


    def override(self, conf: BaselineConf):
        start_hour = conf.start_hour if self.start_hour is None else self.start_hour
        end_hour = conf.end_hour if self.end_hour is None else self.end_hour
        aggregation_period = conf.aggregation_period if self.aggregation_period is None else \
            RecurringMinutes(self.aggregation_period)

        minimums = {self.only_gas: conf.minimum(self.only_gas)} if self.only_gas else conf.minimums

        conf = BaselineConf(conf.name, conf.timezone, start_hour, end_hour, aggregation_period, minimums)

        if self.minimum_gas is not None:
            conf.set_minimum(self.minimum_gas, self.minimum_value)

        if self.exclude_gas is not None:
            conf.remove_minimum(self.exclude_gas)

        return conf


    def excludes_gas(self, gas):
        return self.only_gas is not None and gas != self.only_gas


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def aws(self):
        return self.__opts.aws


    @property
    def conf_name(self):
        return self.__opts.conf_name


    @property
    def fields(self):
        return self.__opts.fields


    @property
    def rehearse(self):
        return self.__opts.rehearse


    @property
    def uptake(self):
        return self.__opts.uptake


    @property
    def start_hour(self):
        return self.__opts.start_hour


    @property
    def end_hour(self):
        return self.__opts.end_hour


    @property
    def aggregation_period(self):
        return self.__opts.aggregation_period


    @property
    def minimum_gas(self):
        return None if self.__opts.minimum is None else self.__opts.minimum[0]


    @property
    def minimum_value(self):
        return None if self.__opts.minimum is None else int(self.__opts.minimum[1])


    @property
    def only_gas(self):
        return self.__opts.only_gas


    @property
    def exclude_gas(self):
        return self.__opts.exclude_gas


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def device_tags(self):
        return self.__args


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdBaseline:{aws:%s, conf_name:%s, fields:%s, rehearse:%s, uptake:%s, start_hour:%s, " \
               "end_hour:%s, aggregation_period:%s, minimum:%s, only_gas:%s, exclude_gas:%s, " \
               "verbose:%s, device_tags:%s}" % \
               (self.aws, self.conf_name, self.fields, self.rehearse, self.uptake, self.start_hour,
                self.end_hour, self.aggregation_period, self.__opts.minimum, self.only_gas, self.exclude_gas,
                self.verbose, self.device_tags)



#!/usr/bin/env python3

"""
Created on 29 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The alert utility is used to create, update, delete and find alert specifications.

Alerts take the form of emails, sent when a parameter falls below or above specified bounds, or when the value is null
(a null value is being reported, or no reports are available). The alert specification sets these bounds, together with
the aggregartion period (usually in minutes). The minimum period is one minute.

In --find mode, results can be filtered by topic, field or creator email address.

SYNOPSIS
alert.py  { -F | -R ID | -C | -U ID | -D ID } [-p TOPIC] [-f FIELD] [-l LOWER] [-u UPPER] [-n { 1 | 0 }]
[-a INTERVAL UNITS] [-t INTERVAL] [-s { 1 | 0 }] [-c EMAIL_ADDR] [-e EMAIL_ADDR] [-i INDENT] [-v]

EXAMPLES
alert.py -v -C -p south-coast-science-dev/development/loc/1/gases -f val.NO2.cnc -u 1000.0 -a 1 M \
-e someone@me.com -i4

DOCUMENT EXAMPLE
{"id": 77, "topic": "south-coast-science-dev/development/loc/1/gases", "field": "val.CO.cnc", "lower-threshold": null,
"upper-threshold": 1000.0, "alert-on-none": true, "aggregation-period": {"interval": 1, "units": "M"},
"test-interval": null, "creator-email-address": "authorization@southcoastscience.com",
"to": "someone@me.com", "cc-list": [], "suspended": false}

SEE ALSO
scs_analysis/alert_status
scs_analysis/monitor_auth

BUGS
* The --test-interval flag is not currently in use, and is ignored.
* Email CC addresses cannot be added or removed.
"""

import json

import requests
import sys

from scs_analysis.cmd.cmd_alert import CmdAlert

from scs_core.aws.client.api_auth import APIAuth
from scs_core.aws.client.monitor_auth import MonitorAuth

from scs_core.aws.data.alert import AlertSpecification

from scs_core.aws.manager.alert_specification_manager import AlertSpecificationManager
from scs_core.aws.manager.byline_manager import BylineManager

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict
from scs_core.data.str import Str

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    auth = None
    response = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdAlert()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('alert', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # MonitorAuth...
        if not MonitorAuth.exists(Host):
            logger.error('MonitorAuth not available.')
            exit(1)

        try:
            auth = MonitorAuth.load(Host, encryption_key=MonitorAuth.password_from_user())
        except (KeyError, ValueError):
            logger.error('incorrect password.')
            exit(1)

        alert_manager = AlertSpecificationManager(requests, auth)
        logger.info(alert_manager)

        # APIAuth...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            logger.error("APIAuth not available.")
            exit(1)

        # byline manager...
        byline_manager = BylineManager(api_auth)
        logger.info(byline_manager)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.find:
            creator_filter = auth.email_address if cmd.creator is None else cmd.creator
            response = alert_manager.find(cmd.topic, cmd.field, creator_filter)
            report = sorted(response.alerts)

        if cmd.retrieve:
            report = alert_manager.retrieve(cmd.retrieve_id)

        if cmd.create:
            # validate...
            if not cmd.is_complete():
                logger.error("minimum parameters are topic, path, a threshold, and an aggregation period.")
                exit(2)

            byline = byline_manager.find_latest_byline_for_topic(cmd.topic)

            if byline is None or cmd.topic != byline.topic:
                logger.error("the topic '%s' is not available." % cmd.topic)
                exit(2)

            message = PathDict(json.loads(byline.message))

            if not message.has_path(cmd.field):
                paths = Str.collection(message.paths())
                logger.error("the field '%s' is not available. Available fields are: %s" % (cmd.field, paths))
                exit(2)

            # create...
            to = auth.email_address if cmd.to is None else cmd.to

            alert = AlertSpecification(None, cmd.topic, cmd.field, cmd.lower_threshold, cmd.upper_threshold,
                                       cmd.alert_on_none, cmd.aggregation_period, cmd.test_interval, auth.email_address,
                                       to, [], cmd.suspended)

            if not alert.has_valid_thresholds():
                logger.error("threshold values are invalid.")
                exit(2)

            if not alert.has_valid_aggregation_period():
                logger.error("the aggregation period is invalid.")
                exit(2)

            report = alert_manager.create(alert)

        if cmd.update:
            # validate...
            if cmd.topic is not None or cmd.field is not None:
                logger.error("topic and field may not be changed.")
                exit(2)

            alert = alert_manager.retrieve(cmd.update_id)

            if alert is None:
                logger.error("no alert found with ID %s." % cmd.update_id)
                exit(2)

            if auth.email_address != alert.creator_email_address:
                logger.error("you do not have permission to update this alert.")
                exit(2)

            # update...
            lower_threshold = alert.lower_threshold if cmd.lower_threshold is None else cmd.lower_threshold
            upper_threshold = alert.upper_threshold if cmd.upper_threshold is None else cmd.upper_threshold
            alert_on_none = alert.alert_on_none if cmd.alert_on_none is None else bool(cmd.alert_on_none)
            aggregation_period = alert.aggregation_period if cmd.aggregation_period is None else cmd.aggregation_period
            test_interval = alert.test_interval if cmd.test_interval is None else cmd.test_interval
            suspended = alert.suspended if cmd.suspended is None else bool(cmd.suspended)
            to = alert.to if cmd.to is None else cmd.to

            updated = AlertSpecification(alert.id, alert.topic, alert.field, lower_threshold, upper_threshold,
                                         alert_on_none, aggregation_period, test_interval, alert.creator_email_address,
                                         to, alert.cc_list, suspended)

            if not updated.has_valid_thresholds():
                logger.error("threshold values are invalid.")
                exit(2)

            if not updated.has_valid_aggregation_period():
                logger.error("the aggregation period is invalid.")
                exit(2)

            report = alert_manager.update(updated)

        if cmd.delete:
            alert = alert_manager.retrieve(cmd.delete_id)

            if alert is None:
                logger.error("no alert found with ID %s." % cmd.delete_id)
                exit(2)

            alert_manager.delete(cmd.delete_id)

        # report...
        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

        if cmd.find:
            logger.info('retrieved: %s' % len(response.alerts))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        now = LocalizedDatetime.now().utc().as_iso8601()
        logger.error("%s: HTTP response: %s (%s) %s" % (now, ex.status, ex.reason, ex.data))
        exit(1)



#!/usr/bin/env python3

"""
Created on 18 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The socket_receiver utility is used to accept data via a Unix socket, with data sourced from the same host, or
another host on the same local area network. A socket_sender utility is provided for the purpose of sourcing data,
as part of the scs_dev package.

The socket_receiver utility should be started before socket_sender. When socket_sender terminates, socket_receiver
will also terminate.

If a port number is not specified, then port 2000 is used.

SYNOPSIS
socket_receiver.py [-p PORT] [-v]

EXAMPLES
socket_receiver.py -p 2002

SEE ALSO
scs_analysis/uds_receiver
scs_dev/socket_sender

BUGS
It is possible to create scenarios where a port becomes orphaned. Depending on host operating systems, orphaned ports
may take time to be garbage collected.
"""

import sys

from scs_analysis.cmd.cmd_socket_receiver import CmdSocketReceiver

from scs_host.comms.network_socket import NetworkSocket


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSocketReceiver()

    if cmd.verbose:
        print("socket_receiver: %s" % cmd, file=sys.stderr)

    receiver = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        receiver = NetworkSocket(port=cmd.port)

        if cmd.verbose:
            print("socket_receiver: %s" % receiver, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        for message in receiver.read():
            print(message)
            sys.stdout.flush()

            receiver.ack()


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)


    # ----------------------------------------------------------------------------------------------------------------
    # close...

    finally:
        if receiver:
            receiver.close()



#!/usr/bin/env python3

"""
Created on 25 Dec 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The aws_byline utility is used to find the date / time of the most-recently published message for a given topic
or device. The user may specify a topic path (find all devices that have published to the given topic), or a device tag
(find all topics which the given device has published to), but not both. A further option --all reports all bylines.

Output is in the form of zero or more JSON documents, indicating the device, topic and localised date / time for each
latest sense event.

SYNOPSIS
aws_byline.py { -d DEVICE | -t TOPIC [-l] | -a } [-x EXCLUDED] [-m] [-v]

EXAMPLES
aws_byline.py -t south-coast-science-demo -v -x /control

DOCUMENT EXAMPLE - OUTPUT
{"device": "scs-bgx-401", "topic": "south-coast-science-demo/brighton/loc/1/climate",
"lastSeenTime": "2020-10-23T08:52:20Z", "last_write": "2020-10-23T08:52:20Z",
"message": "{\"val\": {\"hmd\": 68.4, \"tmp\": 19.8, \"bar\": null}, \"rec\": \"2020-10-23T08:52:20Z\",
\"tag\": \"scs-bgx-401\"}"}

SEE ALSO
scs_analysis/aws_topic_history
"""

import sys

from scs_analysis.cmd.cmd_aws_byline import CmdAWSByline

from scs_core.aws.client.api_auth import APIAuth
from scs_core.aws.manager.byline_manager import BylineManager

from scs_core.client.network import Network
from scs_core.client.resource_unavailable_exception import ResourceUnavailableException

from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    group = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdAWSByline()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("aws_byline: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # APIAuth...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            print("aws_byline: APIAuth not available.", file=sys.stderr)
            exit(1)

        # BylineManager...
        manager = BylineManager(api_auth)

        if cmd.verbose:
            print("aws_byline: %s" % manager, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # check...

        if not Network.is_available():
            if cmd.verbose:
                print("aws_byline: waiting for network.", file=sys.stderr)
                sys.stderr.flush()

            Network.wait()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        latest = None

        # find...
        if cmd.topic:
            group = manager.find_bylines_for_topic(cmd.topic, excluded=cmd.excluded)

        elif cmd.device:
            group = manager.find_bylines_for_device(cmd.device, excluded=cmd.excluded)

        else:
            group = manager.find_bylines(excluded=cmd.excluded)     # all

        # report...
        for byline in group.bylines:
            if cmd.latest:
                if latest is None or latest.rec < byline.rec:
                    latest = byline

            else:
                jdict = byline.as_json(include_message=cmd.include_messages)
                print(JSONify.dumps(jdict))
                sys.stdout.flush()

        if cmd.latest and latest is not None:
            jdict = latest.as_json(include_message=cmd.include_messages)
            print(JSONify.dumps(jdict))


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except (ConnectionError, HTTPException) as ex:
        print("aws_byline: %s: %s" % (ex.__class__.__name__, ex), file=sys.stderr)
        exit(1)

    except ResourceUnavailableException as ex:
        print("aws_byline: %s" % repr(ex), file=sys.stderr)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose and group is not None and len(group):
            latest_pub = group.latest_pub()
            latest_iso = None if latest_pub is None else latest_pub.as_iso8601()

            print("aws_byline: latest_pub: %s" % latest_iso, file=sys.stderr)



"""
Created on 4 Mar 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdSampleStats(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-t TAG] [-i] [-p PRECISION] [-a] [-r] [-v] "
                                                    "PATH1 [PATH2 .. PATHN]", version="%prog 1.0")

        # input...
        self.__parser.add_option("--tag", "-t", type="string", nargs=1, action="store", default='tag', dest="tag",
                                 help="name of the tag field (default 'tag')")

        # output...
        self.__parser.add_option("--include-tag", "-i", action="store_true", dest="include_tag", default=False,
                                 help="include the device tag")

        self.__parser.add_option("--prec", "-p", type="int", nargs=1, action="store", default=6, dest="precision",
                                 help="precision (default 6 decimal places)")

        self.__parser.add_option("--analytic", "-a", action="store_true", dest="analytic", default=False,
                                 help="analytic output")

        self.__parser.add_option("--rows", "-r", action="store_true", dest="rows", default=False,
                                 help="output results for each path on a separate row")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if not self.paths:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__opts.tag


    @property
    def include_tag(self):
        return self.__opts.include_tag


    @property
    def precision(self):
        return self.__opts.precision


    @property
    def analytic(self):
        return self.__opts.analytic


    @property
    def rows(self):
        return self.__opts.rows


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def paths(self):
        return self.__args


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdSampleStats:{tag:%s, include_tag:%s, precision:%s, analytic:%s, rows:%s, verbose:%s, paths:%s}" % \
               (self.tag, self.include_tag, self.precision, self.analytic, self.rows, self.verbose, self.paths)



"""
Created on 18 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

example document:
{"org-id": "south-coast-science-test-user", "api-key": "9fdfb841-3433-45b8-b223-3f5a283ceb8e"}
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdOSIOAPIAuth(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-s ORG_ID API_KEY] [-v]", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--set", "-s", type="string", nargs=2, action="store", dest="org_key",
                                 help="set org ID and API key")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def set(self):
        return self.__opts.org_key is not None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def org_id(self):
        return None if self.__opts.org_key is None else self.__opts.org_key[0]


    @property
    def api_key(self):
        return None if self.__opts.org_key is None else self.__opts.org_key[1]


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CmdOSIOAPIAuth:{org_id:%s, api_key:%s, verbose:%s}" % (self.org_id, self.api_key, self.verbose)



#!/usr/bin/env python3

"""
Created on 3 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_duplicates utility is used to find duplicate values in a sequence of input JSON documents, for a specified
leaf node path. It is particularly useful in searching for duplicate recording datetimes.

If an input document does not contain the specified path, then it is ignored.

The default output report is sequence of JSON dictionaries with a field for each value where duplicates were found.
The value for this field is itself a dictionary, with a field for the index of each input document, and value being
the input document itself.

In the --counts mode, the output report is sequence of JSON dictionaries with a field for each value where duplicates
were found, whose value is the number of matching documents.

SYNOPSIS
sample_duplicates.py [-c] [-v] PATH

EXAMPLES
csv_reader.py climate.csv | sample_duplicates.py -v val.hmd

DOCUMENT EXAMPLE - OUTPUT
default mode:
{"17.5": {"278728": {"val": {"hmd": 17.5, "tmp": 25.7}, "rec": "2019-02-25T15:28:18Z", "tag": "scs-bgx-303"},
"278731": {"val": {"hmd": 17.5, "tmp": 25.7}, "rec": "2019-02-25T15:31:18Z", "tag": "scs-bgx-303"}}}

counts mode:
{"17.5": 2}
"""

import sys

from scs_analysis.cmd.cmd_sample_duplicates import CmdSampleDuplicates

from scs_core.data.duplicates import Duplicates
from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    dupes = None

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleDuplicates()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_duplicates: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        dupes = Duplicates()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        for line in sys.stdin:
            jstr = line.strip()
            datum = PathDict.construct_from_jstr(jstr)

            if datum is None:
                continue

            document_count += 1

            if not datum.has_path(cmd.path):
                continue

            dupes.test(document_count, datum.node(cmd.path), datum)

            processed_count += 1


        # ------------------------------------------------------------------------------------------------------------
        # report...

        if cmd.counts:
            for count in dupes.match_counts():
                print(JSONify.dumps(count))

        else:
            for match in dupes.matches():
                print(JSONify.dumps(match))


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_duplicates: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)

            if dupes is not None:
                print("sample_duplicates: values with duplicates: %d" % dupes.matched_key_count, file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 22 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The cognito_user_credentials utility is used to manage the AWS Cognito credentials for the user. The credentials are
composed of an email address and a password. The JSON identity document managed by this utility is encrypted.

The password must be specified when the credentials are created and is required when the credentials are accessed.

SYNOPSIS
cognito_user_credentials.py [-c CREDENTIALS] [{ -s | -t | -d }] [-v]

EXAMPLES
./cognito_user_credentials.py -s

FILES
~/SCS/aws/cognito_user_credentials.json

DOCUMENT EXAMPLE
{"email": "bruno.beloff@southcoastscience.com", "password": "hello"}

SEE ALSO
scs_analysis/cognito_identity

RESOURCES
https://stackoverflow.com/questions/42568262/how-to-encrypt-text-with-a-password-in-python
"""

import requests
import sys

from scs_analysis.cmd.cmd_cognito_user_credentials import CmdCognitoUserCredentials

from scs_core.aws.security.cognito_login_manager import CognitoUserLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials

from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

def load_credentials(credentials_name):
    try:
        password = CognitoUserCredentials.password_from_user()
        return CognitoUserCredentials.load(Host, name=credentials_name, encryption_key=password)

    except (KeyError, ValueError):
        logger.error("incorrect password")
        exit(1)


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    credentials = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdCognitoUserCredentials()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('cognito_user_credentials', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.set:
            credentials = CognitoUserCredentials.from_user(cmd.credentials_name)

            if not credentials.ok():
                logger.error("the identity is not valid")
                exit(1)

            credentials.save(Host, encryption_key=credentials.password)

        elif cmd.test:
            manager = CognitoUserLoginManager(requests)

            credentials = load_credentials(cmd.credentials_name)

            if credentials is None:
                logger.error("no credentials are available")
                exit(1)

            logger.info(credentials)

            auth = manager.login(credentials)
            logger.info(auth)

            if auth is None:
                logger.error("invalid auth")
                exit(1)

        elif cmd.delete:
            CognitoUserCredentials.delete(Host)

        else:
            credentials = load_credentials(cmd.credentials_name)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if credentials:
            print(JSONify.dumps(credentials))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        logger.error(ex.data)
        exit(1)



"""
Created on 23 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class CmdMQTTClient(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-p UDS_PUB] "
                                                    "[-s] { -c { C | G | P | S | X } (UDS_SUB_1) | "
                                                    "[SUB_TOPIC_1 (UDS_SUB_1) .. SUB_TOPIC_N (UDS_SUB_N)] } "
                                                    "[-n] [-t] [-e] [-v]", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--pub", "-p", type="string", nargs=1, action="store", dest="uds_pub",
                                 default=None, help="read publications from UDS instead of stdin")

        self.__parser.add_option("--sub", "-s", action="store_true", dest="uds_sub",
                                 help="write subscriptions to UDS instead of stdout")

        self.__parser.add_option("--channel", "-c", type="string", nargs=1, action="store", dest="channel",
                                 help="subscribe to channel")

        self.__parser.add_option("--no-wrapper", "-n", action="store_true", dest="no_wrapper", default=False,
                                 help="discard topic wrapper")

        self.__parser.add_option("--timed", "-t", action="store_true", dest="timed", default=False,
                                 help="add a field for received datetime")

        self.__parser.add_option("--echo", "-e", action="store_true", dest="echo", default=False,
                                 help="echo input to stdout (if not writing subscriptions to stdout)")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.echo and self.subscriptions and not self.__opts.uds_sub:
            return False

        if self.channel is None:
            if self.__opts.uds_sub and len(self.__args) % 2 != 0:
                return False

        # else:
        #     if self.__opts.uds_sub and len(self.__args) != 1:
        #         return False
        #
        #     if not self.__opts.uds_sub and len(self.__args) != 0:
        #         return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def subscriptions(self):
        subscriptions = []

        if self.channel:
            return subscriptions

        if self.__opts.uds_sub:
            for i in range(0, len(self.__args), 2):
                subscriptions.append(Subscription(self.__args[i], self.__args[i + 1]))
        else:
            for i in range(len(self.__args)):
                subscriptions.append(Subscription(self.__args[i]))

        return subscriptions


    @property
    def channel(self):
        return self.__opts.channel


    @property
    def channel_uds(self):
        if self.channel is None or not self.__opts.uds_sub:
            return None

        return self.__args[0]


    @property
    def uds_pub(self):
        return self.__opts.uds_pub


    @property
    def no_wrapper(self):
        return self.__opts.no_wrapper


    @property
    def timed(self):
        return self.__opts.timed


    @property
    def wrap(self):
        return not self.__opts.no_wrapper


    @property
    def echo(self):
        return self.__opts.echo


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdMQTTClient:{subscriptions:%s, channel:%s, channel_uds:%s, uds_pub:%s, " \
               "timed:%s, no_wrapper:%s, echo:%s, verbose:%s}" % \
               (Str.collection(self.subscriptions), self.channel, self.channel_uds, self.uds_pub,
                self.timed, self.no_wrapper, self.echo, self.verbose)


# --------------------------------------------------------------------------------------------------------------------

class Subscription(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, address=None):
        """
        Constructor
        """
        self.__topic = topic            # string        topic path
        self.__address = address        # string        DomainSocket address


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def address(self):
        return self.__address


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Subscription:{topic:%s, address:%s}" % (self.topic, self.address)



"""
Created on 19 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdOSIOClientAuth(object):
    """
    unix command line handler
    """

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-u USER_ID] [-d DESCRIPTION] [-v]", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--user", "-u", type="string", nargs=1, action="store", dest="user_id",
                                 help="set user-id (only if device has not yet been registered)")

        self.__parser.add_option("--desc", "-d", type="string", nargs=1, action="store", dest="description",
                                 help="set optional device description")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_complete(self):
        if self.user_id is None:
            return False

        return True


    def set(self):
        if self.__opts.user_id is None and self.__opts.description is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def user_id(self):
        return self.__opts.user_id


    @property
    def description(self):
        return self.__opts.description


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdOSIOClientAuth:{user_id:%s, description:%s, verbose:%s}" % \
               (self.user_id, self.description, self.verbose)



#!/usr/bin/env python3

"""
Created on 20 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The configuration_monitor_check utility is used to report on the configuration monitor's attempt to access all of
the devices known to the system. Status levels are:

* NOR - NO RESPONSE
* ERR - ERROR
* M - MALFORMED:
* MSA - MALFORMED:SAMPLE
* MCO - MALFORMED:CONFIG
* NSP - NOT SUPPORTED
* R - RECEIVED:
* RNW - RECEIVED:NEW
* RUN - RECEIVED:UNCHANGED
* RUP - RECEIVED:UPDATED

SYNOPSIS
configuration_monitor_check.py { -c TAG | [-t TAG [-e]] [-r RESULT] [-o] } [-i INDENT] [-v]

EXAMPLES
configuration_monitor_check.py -r ERR | node.py -s | csv_writer.py

DOCUMENT EXAMPLE
{"tag": "scs-ph1-26", "rec": "2021-05-18T14:36:00Z", "result": "ERROR",
"context": ["TimeoutExpired(['./configuration'], 30)"]}

SEE ALSO
scs_analysis/configuration_csv
scs_analysis/configuration_monitor
scs_analysis/monitor_auth

scs_mfr/configuration
"""

import requests
import sys

from scs_analysis.cmd.cmd_configuration_monitor_check import CmdConfigurationMonitorCheck

from scs_core.aws.client.monitor_auth import MonitorAuth
from scs_core.aws.manager.configuration_check_finder import ConfigurationCheckFinder
from scs_core.aws.manager.configuration_check_requester import ConfigurationCheckRequester

from scs_core.data.json import JSONify
from scs_core.data.datetime import LocalizedDatetime

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    auth = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdConfigurationMonitorCheck()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('configuration_monitor_check', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if not MonitorAuth.exists(Host):
            logger.error('MonitorAuth not available.')
            exit(1)

        try:
            auth = MonitorAuth.load(Host, encryption_key=MonitorAuth.password_from_user())
        except (KeyError, ValueError):
            logger.error('incorrect password.')
            exit(1)

        finder = ConfigurationCheckFinder(requests, auth)
        requester = ConfigurationCheckRequester(requests, auth)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.check_tag:
            response = requester.request(cmd.check_tag)
            print(response.result, file=sys.stderr)
            exit(0 if response.result == 'OK' else 1)

        response = finder.find(cmd.tag_filter, cmd.exact_match, cmd.result_code, cmd.response_mode())
        print(JSONify.dumps(sorted(response.items), indent=cmd.indent))
        logger.info('retrieved: %s' % len(response.items))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        now = LocalizedDatetime.now().utc().as_iso8601()
        logger.error("%s: HTTP response: %s (%s) %s" % (now, ex.status, ex.reason, ex.data))
        exit(1)



#!/usr/bin/env python3

"""
Created on 20 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The osio_topic_history utility is used to retrieve stored data from the OpenSensors.io Community Edition data
infrastructure. Data can be retrieved by start or start + end localised date / times, or by minutes backwards from now.

An OpenSensors.io API auth document must be installed on the host for the osio_mqtt_client
to operate. A specification should be obtained from the user's OpenSensors.io account.

SYNOPSIS
osio_topic_history.py { -m MINUTES | -s START [-e END] } [-p SECONDS] [-w] [-v] PATH

EXAMPLES
osio_topic_history.py -v /orgs/south-coast-science-dev/exhibition/loc/1/particulates -m1

FILES
~/SCS/osio/osio_api_auth.json

SEE ALSO
scs_analysis/osio_api_auth
"""

import sys

from scs_analysis.cmd.cmd_osio_topic_history import CmdOSIOTopicHistory

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONify

from scs_core.osio.client.api_auth import APIAuth
from scs_core.osio.manager.topic_manager import TopicManager
from scs_core.osio.manager.message_manager import MessageManager

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    agent = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdOSIOTopicHistory()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("osio_topic_history: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # APIAuth...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            print("osio_topic_history: APIAuth not available.", file=sys.stderr)
            exit(1)

        if cmd.verbose:
            print("osio_topic_history: %s" % api_auth, file=sys.stderr)
            sys.stderr.flush()

        # topic manager...
        topic_manager = TopicManager(api_auth.api_key)

        # message manager...
        message_manager = MessageManager(api_auth.api_key, verbose=cmd.verbose)

        if cmd.verbose:
            print("osio_topic_history: %s" % message_manager, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # check topics...
        if not topic_manager.find(cmd.path):
            print("osio_topic_history: Topic not available: %s" % cmd.path, file=sys.stderr)
            exit(1)

        # time...
        if cmd.use_offset():
            end = LocalizedDatetime.now()
            start = end.timedelta(minutes=-cmd.minutes)
        else:
            end = LocalizedDatetime.now() if cmd.end is None else cmd.end
            start = cmd.start

        if cmd.verbose:
            print("osio_topic_history: start: %s" % start, file=sys.stderr)
            print("osio_topic_history: end: %s" % end, file=sys.stderr)
            sys.stderr.flush()

        # messages...
        messages = message_manager.find_for_topic(cmd.path, start, end, cmd.pause)

        for message in messages:
            document = message if cmd.include_wrapping else message.payload.content
            print(JSONify.dumps(document))
            sys.stdout.flush()

        if cmd.verbose:
            print("osio_topic_history: total: %d" % len(messages), file=sys.stderr)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)



"""
Created on 17 Aug 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

from subprocess import Popen, PIPE

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCSVGenerator(object):
    """
    classdocs
    """

    COMMON_NODES = ['tag', 'rec', 'rec.report', 'rec.update', 'ver']

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, verbose):
        self.__csv_opts = '-vsq' if verbose else '-sq'
        self.__node_opts = '-v' if verbose else ''


    # ----------------------------------------------------------------------------------------------------------------

    def generate(self, selected_configs, selected_nodes, file_path):
        jstr = '\n'.join([JSONify.dumps(config) for config in selected_configs])
        node_args = self.COMMON_NODES + ['val.' + selected_node for selected_node in selected_nodes]

        if selected_nodes:
            p = Popen(['node.py', self.__node_opts] + node_args, stdin=PIPE, stdout=PIPE)
            csv_data, _ = p.communicate(input=jstr.encode())
        else:
            csv_data = jstr.encode()

        p = Popen(['csv_writer.py', self.__csv_opts, file_path], stdin=PIPE)
        p.communicate(input=csv_data)



"""
Created on 20 Feb 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdCSVCollator(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog -l LOWER_BOUND -u UPPER_BOUND -d DELTA [-f FILE_PREFIX] "
                                                    "[-v] PATH", version="%prog 1.0")

        # compulsory...
        self.__parser.add_option("--lower", "-l", type="float", nargs=1, action="store", dest="lower",
                                 help="lower bound of dataset")

        self.__parser.add_option("--upper", "-u", type="float", nargs=1, action="store", dest="upper",
                                 help="upper bound of dataset")

        self.__parser.add_option("--delta", "-d", type="float", nargs=1, action="store", dest="delta",
                                 help="width of bin")

        # optional...
        self.__parser.add_option("--file-prefix", "-f", type="string", nargs=1, action="store", dest="file_prefix",
                                 help="file prefix for collated CSVs")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.lower is None or self.upper is None or self.delta is None or self.path is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def lower(self):
        return self.__opts.lower


    @property
    def upper(self):
        return self.__opts.upper


    @property
    def delta(self):
        return self.__opts.delta


    @property
    def file_prefix(self):
        return self.__opts.file_prefix


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdCSVCollator:{lower:%s, upper:%s, delta:%s, file_prefix:%s, verbose:%s, path:%s}" % \
               (self.lower, self.upper, self.delta, self.file_prefix, self.verbose, self.path)



#!/usr/bin/env python3

"""
Created on 9 May 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The osio_mqtt_control utility is used to interact with a remote host, using the device's control topic - its
command / receipt message regime provides an interactive system over the OpenSensors.io Community Edition
messaging infrastructure.

For security reasons, the user must be in possession of the target device's unique tag and shared secret.

OpenSensors.io API auth and client specifications must be installed on the host for the osio_mqtt_client
to operate. A specification should be obtained from the user's OpenSensors.io account.

SYNOPSIS
osio_mqtt_control.py -d TAG SHARED_SECRET TOPIC { -i | -r CMD } [-t TIMEOUT] [-v]

EXAMPLES
osio_mqtt_control.py -r -d scs-ap1-6 00000000eda1f8a9 \
 /orgs/south-coast-science-dev/development/device/alpha-pi-eng-000006/control -i

FILES
~/SCS/osio/osio_api_auth.json
~/SCS/osio/osio_client_auth.json

SEE ALSO
scs_analysis/mqtt_control_auth
scs_analysis/osio_api_auth
scs_analysis/osio_client_auth
scs_analysis/osio_mqtt_client
"""

import random
import sys
import time

from scs_analysis.cmd.cmd_mqtt_control import CmdMQTTControl
from scs_analysis.handler.osio_mqtt_control_handler import OSIOMQTTControlHandler

from scs_core.control.control_datum import ControlDatum

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONify
from scs_core.data.publication import Publication

from scs_core.osio.client.api_auth import APIAuth
from scs_core.osio.client.client_auth import ClientAuth
from scs_core.osio.manager.topic_manager import TopicManager

from scs_core.sys.exception_report import ExceptionReport

from scs_host.client.mqtt_client import MQTTClient
from scs_host.client.mqtt_client import MQTTSubscriber

from scs_host.comms.stdio import StdIO
from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    client = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdMQTTControl()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("osio_mqtt_control: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # APIAuth...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            print("osio_mqtt_control: APIAuth not available.", file=sys.stderr)
            exit(1)

        if cmd.verbose:
            print("osio_mqtt_control: %s" % api_auth, file=sys.stderr)

        # ClientAuth...
        client_auth = ClientAuth.load(Host)

        if client_auth is None:
            print("osio_mqtt_control: ClientAuth not available.", file=sys.stderr)
            exit(1)

        if cmd.verbose:
            print("osio_mqtt_control: %s" % client_auth, file=sys.stderr)
            sys.stderr.flush()

        # manager...
        manager = TopicManager(api_auth.api_key)

        # check topic...
        if not manager.find(cmd.device_topic):
            print("osio_mqtt_control: Topic not available: %s" % cmd.device_topic, file=sys.stderr)
            exit(1)

        # responder...
        handler = OSIOMQTTControlHandler()
        subscriber = MQTTSubscriber(cmd.device_topic, handler.handle)

        # client...
        client = MQTTClient(subscriber)

        if cmd.verbose:
            print("osio_mqtt_control: %s" % client, file=sys.stderr)
            sys.stderr.flush()

        # tag...
        tag = Host.name()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        client.connect(ClientAuth.MQTT_HOST, client_auth.client_id, client_auth.user_id, client_auth.client_password)

        while True:
            # cmd...
            if cmd.interactive:
                line = StdIO.prompt(cmd.device_tag + ' > ')
                cmd_tokens = line.split() if len(line) > 0 else None

            else:
                cmd_tokens = cmd.cmd_tokens

            # datum...
            now = LocalizedDatetime.now().utc()
            datum = ControlDatum.construct(tag, cmd.device_tag, now, cmd_tokens, cmd.timeout, cmd.device_shared_secret)

            publication = Publication(cmd.device_topic, datum)
            handler.set(publication)

            if cmd.verbose:
                print(datum, file=sys.stderr)
                sys.stderr.flush()

            # publish...
            while True:
                try:
                    success = client.publish(publication, ClientAuth.MQTT_TIMEOUT)

                    if not success:
                        print("osio_mqtt_control: abandoned", file=sys.stderr)
                        sys.stderr.flush()

                    break

                except Exception as ex:
                    print(JSONify.dumps(ExceptionReport.construct(ex)), file=sys.stderr)
                    sys.stderr.flush()

                time.sleep(random.uniform(1.0, 2.0))

            # subscribe...
            timeout = time.time() + cmd.timeout

            if cmd.receipt or cmd.interactive:
                while True:
                    if handler.receipt:
                        if not handler.receipt.is_valid(cmd.device_shared_secret):
                            raise ValueError("osio_mqtt_control: invalid digest: %s" % handler.receipt)

                        if cmd.verbose:
                            print(handler.receipt, file=sys.stderr)

                        if handler.receipt.command.stderr:
                            print(*handler.receipt.command.stderr, sep='\n', file=sys.stderr)

                        if handler.receipt.command.stdout:
                            print(*handler.receipt.command.stdout, sep='\n')

                        break

                    if cmd.interactive and time.time() > timeout:
                        break

                    time.sleep(0.1)

            if not cmd.interactive:
                break


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if client:
            client.disconnect()



#!/usr/bin/env python3

"""
Created on 19 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The organisation_devices utility is used to

SYNOPSIS
organisation_devices.py [-c CREDENTIALS] { -F { -l ORG_LABEL | -t DEVICE_TAG } | \
-C -l ORG_LABEL -t DEVICE_TAG -p PATH_ROOT GROUP LOCATION -d DEPLOYMENT_LABEL | \
-E -l ORG_LABEL -t DEVICE_TAG -p PATH_ROOT GROUP LOCATION | \
-D -t DEVICE_TAG } \
[-i INDENT] [-v]

EXAMPLES
organisation_devices.py -F -l NARA

DOCUMENT EXAMPLE
{"DeviceTag": "scs-bgx-401", "OrgID": 1, "DevicePath": "south-coast-science-demo/brighton/loc/1/",
"EnvironmentPath": "south-coast-science-demo/brighton/device/praxis-000401/",
"StartDatetime": "2022-01-17T10:40:04Z", "EndDatetime": null,
"DeploymentLabel": "Preston Circus"}

SEE ALSO
scs_analysis/cognito_credentials
"""

import requests
import sys

from scs_analysis.cmd.cmd_organisation_devices import CmdOrganisationDevices

from scs_core.aws.config.project import Project

from scs_core.aws.security.cognito_login_manager import CognitoUserLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials

from scs_core.aws.security.organisation import Organisation, OrganisationPathRoot, OrganisationDevice
from scs_core.aws.security.organisation_manager import OrganisationManager

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# TODO: organisation_devices: {'message': 'Missing Authentication Token'}
# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    credentials = None
    auth = None
    cognito = None
    org = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdOrganisationDevices()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('organisation_devices', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)

        if cmd.org_label is not None and not Organisation.is_valid_label(cmd.org_label):
            logger.error("the organisation label '%s' is not valid." % cmd.org_label)
            exit(2)

        if cmd.device_tag is not None and not OrganisationDevice.is_valid_tag(cmd.device_tag):
            logger.error("the device tag '%s' is not valid." % cmd.device_tag)
            exit(2)

        if cmd.project_organisation is not None and \
                not OrganisationPathRoot.is_valid_path_root(cmd.project_organisation):
            logger.error("the path root '%s' is not valid." % cmd.project_organisation)
            exit(2)

        if cmd.project_location is not None and not Datum.is_int(cmd.project_location):
            logger.error("the project location '%s' must be an integer." % cmd.project_location)
            exit(2)

        if cmd.deployment_label is not None and not OrganisationDevice.is_valid_deployment_label(cmd.deployment_label):
            logger.error("the deployment label '%s' is not valid." % cmd.deployment_label)
            exit(2)


        # ------------------------------------------------------------------------------------------------------------
        # auth...

        gatekeeper = CognitoUserLoginManager(requests)

        # CognitoUserCredentials...
        if not CognitoUserCredentials.exists(Host, name=cmd.credentials_name):
            logger.error("Cognito credentials not available.")
            exit(1)

        try:
            password = CognitoUserCredentials.password_from_user()
            credentials = CognitoUserCredentials.load(Host, name=cmd.credentials_name, encryption_key=password)
        except (KeyError, ValueError):
            logger.error("incorrect password")
            exit(1)

        try:
            auth = gatekeeper.login(credentials)

        except HTTPException as ex:
            logger.error(ex.data)
            exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        manager = OrganisationManager(requests)


        # ------------------------------------------------------------------------------------------------------------
        # validate...

        if cmd.org_label is not None:
            org = manager.get_organisation_by_label(auth.id_token, cmd.org_label)

            if org is None:
                logger.error("no organisation found for label: '%s'." % cmd.org_label)
                exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.find:
            if cmd.org_label is not None:
                report = manager.find_devices_by_organisation(auth.id_token, org.org_id)
            else:
                report = manager.find_devices_by_tag(auth.id_token, cmd.device_tag)

        if cmd.create:
            project = Project.construct(cmd.project_organisation, cmd.project_group, cmd.project_location)
            device_path = project.device_path + '/'
            location_path = project.location_path + '/'

            now = LocalizedDatetime.now()

            report = OrganisationDevice(cmd.device_tag, org.org_id, device_path, location_path, now, None,
                                        cmd.deployment_label)

            manager.assert_device(auth.id_token, report)

        if cmd.expire:
            project = Project.construct(cmd.project_organisation, cmd.project_group, cmd.project_location)
            device_path = project.device_path + '/'
            location_path = project.location_path + '/'

            report = manager.get_device(auth.id_token, cmd.device_tag, org.org_id,
                                        device_path, location_path)

            if report is None:
                logger.error("no device found.")
                exit(1)

            report.end_datetime = LocalizedDatetime.now()
            manager.assert_device(auth.id_token, report)

        if cmd.delete:
            project = Project.construct(cmd.project_organisation, cmd.project_group, cmd.project_location)
            device_path = project.device_path + '/'
            location_path = project.location_path + '/'

            manager.delete_device(auth.id_token, cmd.device_tag, org.org_id, device_path, location_path)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        logger.error(ex.data)
        exit(1)



"""
Created on 3 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse

from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class CmdSampleTimeShift(object):
    """
    unix command line handler
    """

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog -t { + | - } [[DD-]HH:]MM[:SS] [-v] [PATH]",
                                              version="%prog 1.0")

        # functions...
        self.__parser.add_option("--timedelta", "-t", type="string", nargs=2, action="store", dest="timedelta",
                                 help="sign and offset in days / hours / minutes / seconds")

        # output...
        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.__opts.timedelta is None:
            return False

        if self.__opts.timedelta[0] != '+' and self.__opts.timedelta[0] != '-':
            return False

        if self.timedelta is None:
            return False

        return True


    def is_valid_timedelta(self):
        if self.__opts.timedelta is None:
            return True

        return self.timedelta is not None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def positive(self):
        return None if self.__opts.timedelta is None else self.__opts.timedelta[0] == '+'


    @property
    def timedelta(self):
        return Timedelta.construct_from_flag(self.__opts.timedelta[1])


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else 'rec'


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdSampleTimeShift:{timedelta:%s, verbose:%s, path:%s}" % \
               (self.__opts.timedelta, self.verbose, self.path)



#!/usr/bin/env python3

"""
Created on 25 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_min utility is used to find the record with the lowest value for the given node.

Input data is typically in the form of a sequence of JSON documents. A command parameter specifies the path to the node
within the document that is to be tested. The node is typically a leaf node integer or float. The output of the
sample_min utility includes the whole input document.

If there are multiple input documents with the same maximum value, the first document only is written to stdout.

SYNOPSIS
sample_min.py [-v] [PATH]

EXAMPLES
osio_topic_history.py -m60 /orgs/south-coast-science-demo/brighton/loc/1/gases | sample_min.py val.CO.cnc

DOCUMENT EXAMPLE - INPUT
{"tag": "scs-bgx-401", "rec": "2018-03-27T09:54:41.042+00:00", "val": {
"NO2": {"weV": 0.29563, "aeV": 0.280879, "weC": 0.009569, "cnc": 61.0},
"Ox": {"weV": 0.406819, "aeV": 0.387443, "weC": -0.010706, "cnc": 34.7},
"NO": {"weV": 0.319692, "aeV": 0.292129, "weC": 0.028952, "cnc": 165.5},
"CO": {"weV": 0.395819, "aeV": 0.289317, "weC": 0.113108, "cnc": 311.3},
"sht": {"hmd": 82.4, "tmp": 12.6}}}

DOCUMENT EXAMPLE - OUTPUT
{"tag": "scs-bgx-401", "rec": "2018-03-27T09:54:41.042+00:00", "val": {
"NO2": {"weV": 0.29563, "aeV": 0.280879, "weC": 0.009569, "cnc": 61.0},
"Ox": {"weV": 0.406819, "aeV": 0.387443, "weC": -0.010706, "cnc": 34.7},
"NO": {"weV": 0.319692, "aeV": 0.292129, "weC": 0.028952, "cnc": 165.5},
"CO": {"weV": 0.395819, "aeV": 0.289317, "weC": 0.113108, "cnc": 311.3},
"sht": {"hmd": 82.4, "tmp": 12.6}}}
"""

import sys

from scs_analysis.cmd.cmd_sample_record import CmdSampleRecord

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleRecord()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_min: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()


    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        min_datum = None

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            document_count += 1

            if cmd.path not in datum.paths():
                continue

            try:
                value = float(datum.node(cmd.path))
            except (TypeError, ValueError):
                continue

            if min_datum is None or value < min_datum.node(cmd.path):
                min_datum = datum

            processed_count += 1

        if min_datum:
            print(JSONify.dumps(min_datum))


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("sample_min: KeyError: %s" % ex, file=sys.stderr)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_min: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 24 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The cognito_users utility is used to create, update and retrieve AWS Cognito identities. This utility can only be used
by organisation administrators and superusers.

If the --Create function is used, an email is sent to the new user. The verification link in the email must be
excercised in order for the account to gain a CONFIRMED status.

SYNOPSIS
cognito_devices.py  [-c CREDENTIALS] { -F [-t TAG] | -C TAG SHARED_SECRET | -U TAG SHARED_SECRET | -D TAG }
[-i INDENT] [-v]

EXAMPLES
cognito_users.py -Fe bruno.beloff@southcoastscience.com

DOCUMENT EXAMPLE
{"username": "8", "creation-date": "2021-11-24T12:51:12Z", "confirmation-status": "CONFIRMED", "enabled": true,
"email": "bruno.beloff@southcoastscience.com", "given-name": "Bruno", "family-name": "Beloff", "is-super": true}

SEE ALSO
scs_analysis/cognito_credentials

RESOURCES
https://docs.aws.amazon.com/cognito/latest/developerguide/signing-up-users-in-your-app.html
https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-policies.html
"""

import requests
import sys

from scs_analysis.cmd.cmd_cognito_devices import CmdCognitoDevices

from scs_core.aws.security.cognito_device_manager import CognitoDeviceCreator, CognitoDeviceEditor, CognitoDeviceDeleter

from scs_core.aws.security.cognito_device import CognitoDeviceIdentity
from scs_core.aws.security.cognito_device_finder import CognitoDeviceFinder
from scs_core.aws.security.cognito_login_manager import CognitoUserLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials

from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException, HTTPConflictException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    credentials = None
    auth = None
    finder = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdCognitoDevices()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('cognito_devices', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # auth...

        gatekeeper = CognitoUserLoginManager(requests)

        # CognitoUserCredentials...
        if not CognitoUserCredentials.exists(Host, name=cmd.credentials_name):
            logger.error("Cognito credentials not available.")
            exit(1)

        try:
            password = CognitoUserCredentials.password_from_user()
            credentials = CognitoUserCredentials.load(Host, name=cmd.credentials_name, encryption_key=password)
        except (KeyError, ValueError):
            logger.error("incorrect password")
            exit(1)

        try:
            auth = gatekeeper.login(credentials)

        except HTTPException as ex:
            logger.error(ex.data)
            exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if not cmd.create:
            finder = CognitoDeviceFinder(requests)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.find:
            if cmd.tag is not None:
                report = sorted(finder.find_by_tag(auth.id_token, cmd.tag))

            else:
                report = sorted(finder.find_all(auth.id_token))

        if cmd.create:
            if not CognitoDeviceIdentity.is_valid_password(cmd.create[1]):
                logger.error("password must be at least 16 characters.")
                exit(2)

            # create...
            identity = CognitoDeviceIdentity(cmd.create[0], cmd.create[1], None)

            manager = CognitoDeviceCreator(requests, auth.id_token)
            report = manager.create(identity)

        if cmd.update:
            if not CognitoDeviceIdentity.is_valid_password(cmd.update[1]):
                logger.error("password must be at least 16 characters.")
                exit(2)

            # find...
            device = finder.find_by_tag(auth.id_token, cmd.update[0])

            if device is None:
                logger.error("no device found for tag: '%s'." % cmd.update)
                exit(1)

            # update...
            report = CognitoDeviceIdentity(cmd.update[0], cmd.update[1], device.creation_date)

            auth = gatekeeper.login(credentials)                          # renew credentials
            manager = CognitoDeviceEditor(requests, auth.id_token)
            manager.update(report)

        if cmd.delete:
            # TODO: delete device from organisations
            manager = CognitoDeviceDeleter(requests, auth.id_token)
            manager.delete(cmd.delete)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPConflictException as ex:
        logger.error("the tag '%s' is already in use." % report.tag)
        exit(1)



#!/usr/bin/env python3

"""
Created on 23 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The osio_mqtt_client utility is used to subscribe or publish using the OpenSensors.io Community Edition messaging
infrastructure.

OpenSensors.io API auth and client specifications must be installed on the host for the osio_mqtt_client
to operate. A specification should be obtained from the user's OpenSensors.io account.

Only one MQTT client should run at any one time, per TCP/IP host.

SYNOPSIS
osio_mqtt_client.py [-p UDS_PUB] [-s] [SUB_TOPIC_1 (UDS_SUB_1) .. SUB_TOPIC_N (UDS_SUB_N)] [-e] [-v]

EXAMPLES
osio_mqtt_client.py /orgs/south-coast-science-dev/production-test/loc/1/gases

FILES
~/SCS/osio/osio_api_auth.json
~/SCS/osio/osio_client_auth.json

SEE ALSO
scs_analysis/osio_api_auth
scs_analysis/osio_mqtt_control
scs_analysis/osio_topic_publisher

BUGS
When run as a background process, osio_mqtt_client will exit if it has no stdin stream.
"""

import json
import random
import sys
import time

from scs_analysis.cmd.cmd_mqtt_client import CmdMQTTClient
from scs_analysis.handler.osio_mqtt_client_handler import OSIOMQTTHandler
from scs_analysis.handler.mqtt_reporter import MQTTReporter

from scs_core.data.json import JSONify
from scs_core.data.publication import Publication

from scs_core.osio.client.api_auth import APIAuth
from scs_core.osio.client.client_auth import ClientAuth
from scs_core.osio.manager.topic_manager import TopicManager

from scs_core.sys.exception_report import ExceptionReport

from scs_host.client.mqtt_client import MQTTClient, MQTTSubscriber

from scs_host.comms.domain_socket import DomainSocket
from scs_host.comms.stdio import StdIO

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    client = None
    pub_comms = None


    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdMQTTClient()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("osio_mqtt_client: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # APIAuth...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            print("osio_mqtt_client: APIAuth not available.", file=sys.stderr)
            exit(1)

        if cmd.verbose:
            print("osio_mqtt_client: %s" % api_auth, file=sys.stderr)

        # ClientAuth...
        client_auth = ClientAuth.load(Host)

        if client_auth is None:
            print("osio_mqtt_client: ClientAuth not available.", file=sys.stderr)
            exit(1)

        if cmd.verbose:
            print("osio_mqtt_client: %s" % client_auth, file=sys.stderr)

        # comms...
        pub_comms = StdIO()     # DomainSocket(cmd.uds_pub_addr) if cmd.uds_pub_addr else StdIO()

        # manager...
        manager = TopicManager(api_auth.api_key)

        # check topics...
        unavailable = False
        for subscription in cmd.subscriptions:
            if not manager.find(subscription.topic):
                print("osio_mqtt_client: Topic not available: %s" % subscription[0], file=sys.stderr)
                unavailable = True

        if unavailable:
            exit(1)

        # reporter...
        reporter = MQTTReporter(cmd.verbose)

        # subscribers...
        subscribers = []

        for subscription in cmd.subscriptions:
            sub_comms = DomainSocket(subscription.address) if subscription.address else StdIO()

            # handler...
            handler = OSIOMQTTHandler(reporter, sub_comms, cmd.echo)

            if cmd.verbose:
                print("osio_mqtt_client: %s" % handler, file=sys.stderr)
                sys.stderr.flush()

            subscribers.append(MQTTSubscriber(subscription.topic, handler.handle))

        # client...
        client = MQTTClient(*subscribers)
        client.connect(ClientAuth.MQTT_HOST, client_auth.client_id, client_auth.user_id, client_auth.client_password)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # publish...
        pub_comms.connect()

        for message in pub_comms.read():
            try:
                datum = json.loads(message)
            except ValueError:
                reporter.print("bad datum: %s" % message)
                continue

            success = False

            while True:
                publication = Publication.construct_from_jdict(datum)

                try:
                    success = client.publish(publication, ClientAuth.MQTT_TIMEOUT)

                    if not success:
                        reporter.print("abandoned")

                    break

                except Exception as ex:
                    if cmd.verbose:
                        print(JSONify.dumps(ExceptionReport.construct(ex)))
                        sys.stderr.flush()

                time.sleep(random.uniform(1.0, 2.0))        # Don't hammer the broker!

            if success:
                reporter.print("done")

            if cmd.echo:
                print(message)
                sys.stdout.flush()


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if client:
            client.disconnect()

        if pub_comms:
            pub_comms.close()



#!/usr/bin/env python3

"""
Created on 19 Dec 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The osio_client_auth utility is used to store or read the client ID and client password required by the OpenSensors.io
Community Edition messaging system. This client authentication is required to both subscribe to and publish on any
messaging topic.

When setting the client authentication, the osio_client_auth utility requests a new device identity from the
OpenSensors.io system, then stores the generated tokens on the client. The name of the device is taken to be the
name of the host on which the script is executed. Names (unlike client IDs) are not required to be unique on the
OpenSensors system.

SYNOPSIS
osio_client_auth.py [-u USER_ID] [-d DESCRIPTION] [-v]

EXAMPLES
osio_client_auth.py -u south-coast-science-test-user -v

FILES
~/SCS/osio/osio_client_auth.json

DOCUMENT EXAMPLE
{"user_id": "south-coast-science-test-user", "client-id": "5403", "client-password": "rtxSrK2f"}

SEE ALSO
scs_analysis/osio_api_auth
scs_analysis/osio_mqtt_client
scs_analysis/osio_mqtt_control
"""

import sys

from scs_analysis.cmd.cmd_osio_client_auth import CmdOSIOClientAuth

from scs_core.data.json import JSONify

from scs_core.osio.client.api_auth import APIAuth
from scs_core.osio.client.client_auth import ClientAuth
from scs_core.osio.config.project_client import ProjectClient
from scs_core.osio.manager.device_manager import DeviceManager
from scs_core.osio.manager.user_manager import UserManager

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdOSIOClientAuth()

    if cmd.verbose:
        print("osio_client_auth: %s" % cmd, file=sys.stderr)


    # ----------------------------------------------------------------------------------------------------------------
    # resources...

    # APIAuth...
    api_auth = APIAuth.load(Host)

    if api_auth is None:
        print("osio_client_auth: APIAuth not available.", file=sys.stderr)
        exit(1)

    if cmd.verbose:
        print("osio_client_auth: %s" % api_auth, file=sys.stderr)

    # User manager...
    user_manager = UserManager(api_auth.api_key)

    # Device manager...
    device_manager = DeviceManager(api_auth.api_key)

    # check for existing registration...
    device = device_manager.find_for_name(api_auth.org_id, Host.name())

    # check for existing ClientAuth...
    client_auth = ClientAuth.load(Host)


    # ----------------------------------------------------------------------------------------------------------------
    # remove non-matching record...

    if device and client_auth:
        if device.client_id != client_auth.client_id:
            ClientAuth.delete(Host)
            client_auth = None


    # ----------------------------------------------------------------------------------------------------------------
    # validate...

    if device is None or client_auth is None:
        if cmd.set() and not cmd.is_complete():
            print("osio_client_auth: no auth is registered - you must therefore set a user.", file=sys.stderr)
            cmd.print_help(sys.stderr)
            exit(2)

        if not cmd.set():
            exit(0)


    # ----------------------------------------------------------------------------------------------------------------
    # run...

    if cmd.set():
        # User...
        if cmd.user_id:
            user = user_manager.find_public(cmd.user_id)

            if user is None:
                print("osio_client_auth: User not available.", file=sys.stderr)
                exit(1)

        # tags...
        tags = ProjectClient.tags()

        if device and client_auth:
            if cmd.user_id:
                print("osio_client_auth: Device owner-id cannot be updated.", file=sys.stderr)
                exit(1)

            # update Device...
            updated = ProjectClient.update(device, cmd.description, tags)
            device_manager.update(api_auth.org_id, device.client_id, updated)

            # find updated device...
            device = device_manager.find(api_auth.org_id, device.client_id)

        else:
            # create Device...
            device = ProjectClient.create(Host.name(), api_auth, cmd.description, tags)
            device = device_manager.create(cmd.user_id, device)

            # create ClientAuth...
            client_auth = ClientAuth(cmd.user_id, device.client_id, device.password)

            client_auth.save(Host)

    else:
        # find ClientAuth...
        client_auth = ClientAuth.load(Host)


    if cmd.verbose:
        print(device, file=sys.stderr)

    print(JSONify.dumps(client_auth))



"""
Created on 25 Dec 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdAWSByline(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -d DEVICE | -t TOPIC [-l] | -a } [-x EXCLUDED] [-m] [-v]",
                                              version="%prog 1.0")

        # search types...
        self.__parser.add_option("--device", "-d", type="string", nargs=1, action="store", dest="device",
                                 help="report bylines for DEVICE")

        self.__parser.add_option("--topic", "-t", type="string", nargs=1, action="store", dest="topic",
                                 help="report bylines for TOPIC")

        self.__parser.add_option("--all", "-a", action="store_true", dest="all", default=False,
                                 help="report all bylines")

        # output...
        self.__parser.add_option("--excluded", "-x", type="string", nargs=1, action="store", dest="excluded",
                                 help="exclude topics ending with EXCLUDED")

        self.__parser.add_option("--latest", "-l", action="store_true", dest="latest", default=False,
                                 help="only report the most recent byline")

        self.__parser.add_option("--include-messages", "-m", action="store_true", dest="include_messages",
                                 default=False, help="report the message with each byline")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if bool(self.device):
            count += 1

        if bool(self.topic):
            count += 1

        if self.all:
            count += 1

        if count != 1:
            return False

        if self.latest and not bool(self.topic):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        return self.__opts.device


    @property
    def topic(self):
        return self.__opts.topic


    @property
    def latest(self):
        return self.__opts.latest


    @property
    def all(self):
        return self.__opts.all


    @property
    def excluded(self):
        return self.__opts.excluded


    @property
    def include_messages(self):
        return self.__opts.include_messages


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdAWSByline:{device:%s, topic:%s, latest:%s, all:%s, excluded:%s, include_messages:%s, verbose:%s}" % \
               (self.device, self.topic, self.latest, self.all, self.excluded, self.include_messages, self.verbose)



#!/usr/bin/env python3

"""
Created on 7 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The configuration_monitor utility is used to download configuration information from all devices accessible to the
user, and render this in CSV form. Four modes are available:

* --separate - download the latest configuration for all devices, and save this to separate CSV files, each named
for a node. Any nodes specified on the command line are ignored.

* --latest - download the latest configuration for all devices, and save this in the named CSV file.

* --diff-histories - download the history of configuration changes for all devices. Where there has been at least one
change in the configuration for each device, save the history files in the named directory.

* --full-histories - download the history of all configuration changes for all devices.

Nodes are any of:

hostname, packs, afe-baseline, afe-id, aws-api-auth, aws-client-auth, aws-group-config, aws-project, csv-logger-conf,
display-conf, gas-baseline, gas-model-conf, gps-conf, interface-conf, greengrass-identity, mpl115a2-calib, mqtt-conf,
ndir-conf, opc-conf, pmx-model-conf, pressure-conf, psu-conf, psu-version, pt1000-calib, scd30-baseline, scd30-conf,
schedule, shared-secret, sht-conf, sht-conf, networks, modem, sim, system-id, timezone-conf

For separate and latest modes, two rec values are included:

* rec.report - the most recent time that the device reported its status

* rec.update - the most recent time that at least one of the fields changed value (not that those fields might not
be included in the CSV file)

For diff-histories and full-histories modes, a single rec value is included, equivalent to rec.update.

Output CSV cell values are always wrapped in quotes ('"').

SYNOPSIS
configuration_csv.py { -n | -s | -l OUTPUT_CSV | { -d | -f } [-o OUTPUT_CSV_DIR] } [-t DEVICE_TAG [-e]] [-v]
[NODE_1..NODE_N]

EXAMPLES
configuration_csv.py -vs configs.csv
configuration_csv.py -vdo afe_ids afe-id
configuration_csv.py -vft scs-bgx-431

SEE ALSO
scs_analysis/configuration_monitor
scs_analysis/configuration_monitor_check
scs_analysis/monitor_auth

scs_mfr/configuration
"""

import os
import requests
import sys

from scs_analysis.cmd.cmd_configuration_csv import CmdConfigurationCSV
from scs_analysis.handler.batch_download_reporter import BatchDownloadReporter
from scs_analysis.handler.configuration_csv_generator import ConfigurationCSVGenerator

from scs_core.aws.client.monitor_auth import MonitorAuth
from scs_core.aws.manager.configuration_check_finder import ConfigurationCheckFinder, ConfigurationCheckRequest
from scs_core.aws.manager.configuration_finder import ConfigurationFinder, ConfigurationRequest

from scs_core.data.datetime import LocalizedDatetime

from scs_core.estate.configuration import Configuration

from scs_core.sample.configuration_sample import ConfigurationReport

from scs_core.sys.filesystem import Filesystem
from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    configuration = Configuration.construct_from_jdict(None, skeleton=True)
    node_names = sorted(list(configuration.as_json().keys()))

    logger = None
    auth = None
    configs = []

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdConfigurationCSV()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('configuration_csv', verbose=cmd.verbose)     # verbose=cmd.verbose | level=logging.DEBUG
        logger = Logging.getLogger()

        for node in cmd.nodes:
            if node not in node_names:
                logger.error('nodes must be in: %s' % str(node_names))
                exit(2)

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # nodes...
        if cmd.node_names:
            print(node_names, file=sys.stderr)
            exit(0)

        # MonitorAuth...
        if not MonitorAuth.exists(Host):
            logger.error('MonitorAuth not available.')
            exit(1)

        try:
            auth = MonitorAuth.load(Host, encryption_key=MonitorAuth.password_from_user())
        except (KeyError, ValueError):
            logger.error('incorrect password.')
            exit(1)

        # reporter...
        reporter = BatchDownloadReporter()

        # ConfigurationFinder...
        configuration_finder = ConfigurationFinder(requests, auth, reporter=reporter)
        check_finder = ConfigurationCheckFinder(requests, auth)

        # ConfigurationCSVGenerator...
        csv_generator = ConfigurationCSVGenerator(cmd.verbose)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        logger.info("retrieving check reports...")
        checks = check_finder.find(None, None, None, ConfigurationCheckRequest.Mode.FULL)
        check_reports = {check.tag: check.rec for check in checks.items}

        if cmd.separate or cmd.latest:
            logger.info("retrieving configurations...")
            response = configuration_finder.find(cmd.device_tag, cmd.exact_match, cmd.request_mode())
            configs = [ConfigurationReport.construct(item, check_reports[item.tag]) for item in sorted(response)]

        if cmd.separate:
            for node in node_names:
                path = node + '.csv'
                logger.info("-")
                logger.info(path)

                csv_generator.generate(configs, [node], path)

        if cmd.latest:
            logger.info(cmd.latest)
            csv_generator.generate(configs, cmd.nodes, cmd.latest)

        if cmd.diff_histories or cmd.full_histories:
            if cmd.output_csv_dir is not None:
                Filesystem.mkdir(cmd.output_csv_dir)

            logger.info("retrieving device tags...")
            tag_response = configuration_finder.find(cmd.device_tag, cmd.exact_match,
                                                     ConfigurationRequest.Mode.TAGS_ONLY)

            for tag in sorted(tag_response):
                filename = tag + '-configs.csv'
                path = filename if cmd.output_csv_dir is None else os.path.join(cmd.output_csv_dir, filename)

                logger.info("-")
                logger.info(path)

                response = configuration_finder.find(tag, True, cmd.request_mode())
                configs = sorted(response)

                csv_generator.generate(configs, cmd.nodes, path)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        now = LocalizedDatetime.now().utc().as_iso8601()
        logger.error("%s: HTTP response: %s (%s) %s" % (now, ex.status, ex.reason, ex.data))
        exit(1)



#!/usr/bin/env python3

"""
Created on 8 Oct 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The aws_mqtt_control utility is used to interact with a remote host, using the device's control topic.
A command / receipt message regime provides an interactive system over the messaging infrastructure.

For security reasons, the user must be in possession of the target device's unique tag and shared secret. These
tokens can be supplied on the command line, or sourced from an MQTT control auth document managed by the
mqtt_peer utility.

In order to publish on the MQTT control topic, the host must hold an appropriate authentication certificate.
Certificates are available on request from South Coast Science. The certificate should be indicated using the
aws_client_auth utility.

In the interactive mode, the aws_mqtt_control command-line interpreter supports history [UP] and [DOWN] keys.
The mode supports command completion with the [TAB] key and command listing with [TAB][TAB]. Exit from the interactive
mode with [CTRL-D].

SYNOPSIS
aws_mqtt_control.py { -p HOSTNAME [-a] | -d TAG SHARED_SECRET TOPIC } { -i | -r [CMD_TOKENS] } [-t TIMEOUT] [-v]

EXAMPLES
aws_mqtt_control.py -p scs-bbe-002 -r "disk_usage"

FILES
~/SCS/aws/aws_client_auth.json
~/SCS/aws/aws_mqtt_control_history

SEE ALSO
scs_analysis/aws_client_auth
scs_analysis/aws_mqtt_client
scs_analysis/mqtt_peer
scs_mfr/aws_project
scs_mfr/shared_secret
scs_mfr/system_id

BUGS
If the device is currently uploading a backlog of data, then the control receipt will be held in the device's
message queue, and the interaction will time out.
"""

import json
import os
import sys

from scs_analysis.cmd.cmd_mqtt_control import CmdMQTTControl

from scs_core.aws.client.access_key import AccessKey
from scs_core.aws.client.client import Client
from scs_core.aws.client.client_auth import ClientAuth
from scs_core.aws.client.mqtt_client import MQTTClient, MQTTSubscriber
from scs_core.aws.manager.s3_manager import S3PersistenceManager

from scs_core.control.control_handler import ControlHandler

from scs_core.estate.mqtt_peer import MQTTPeerSet

from scs_core.sys.logging import Logging

from scs_host.comms.stdio import StdIO
from scs_host.sys.host import Host


# TODO: test "configuration" command when using mobile network - remove omd field from communications?
# --------------------------------------------------------------------------------------------------------------------

EXIT_COMMANDS = ['reboot', 'restart', 'shutdown']

# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    history_filename = os.path.join(Host.scs_path(), AccessKey.aws_dir(), 'aws_mqtt_control_history')

    key = None
    mqtt_client = None

    stdout = None
    stderr = None
    return_code = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdMQTTControl()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('aws_mqtt_control', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # tag...
        host_tag = Host.name()

        if cmd.aws:
            if not AccessKey.exists(Host):
                logger.error("AccessKey not available.")
                exit(1)

            try:
                key = AccessKey.load(Host, encryption_key=AccessKey.password_from_user())
            except (KeyError, ValueError):
                logger.error("incorrect password")
                exit(1)

            s3_client = Client.construct('s3', key)
            s3_resource_client = Client.resource('s3', key)

            manager = S3PersistenceManager(s3_client, s3_resource_client)

        else:
            manager = Host

        if cmd.is_stored_peer():
            peer_group = MQTTPeerSet.load(manager)
            peer = peer_group.peer(cmd.peer_hostname)

            if peer is None:
                logger.error("no MQTT peer found for host '%s'." % cmd.peer_hostname)
                exit(2)

            logger.info(peer)

            device_tag = peer.tag
            key = peer.shared_secret
            topic = peer.topic

        else:
            device_tag = cmd.device_tag
            key = cmd.device_shared_secret
            topic = cmd.device_topic

        # ClientAuth...
        auth = ClientAuth.load(Host)

        if auth is None:
            logger.error("ClientAuth not available.")
            exit(1)

        logger.info(auth)

        # responder...
        handler = ControlHandler(host_tag, device_tag)
        subscriber = MQTTSubscriber(topic, handler.handle)

        # client...
        mqtt_client = MQTTClient(subscriber)

        logger.info(mqtt_client)


        # ------------------------------------------------------------------------------------------------------------
        # StdIO settings...

        mqtt_client.connect(auth, False)

        if cmd.interactive:
            try:
                stdout, stderr, _ = handler.publish(mqtt_client, topic, ['?'], cmd.DEFAULT_TIMEOUT, key)
            except TimeoutError:
                logger.error("%s is not available." % device_tag)
                exit(1)

            if stderr:
                logger.error("%s device problem: %s." % (device_tag, stderr[0]))
                exit(1)

            commands = json.loads(stdout[0])
            vocabulary = [command + ' ' for command in commands]

            StdIO.set(vocabulary=vocabulary, history_filename=history_filename)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        while True:
            # cmd...
            if cmd.interactive:
                line = StdIO.prompt(device_tag + ' > ')

                if not line:
                    continue

                cmd_tokens = line.split()

            else:
                cmd_tokens = cmd.cmd_tokens

            # publish...
            try:
                stdout, stderr, return_code = handler.publish(mqtt_client, topic, cmd_tokens, cmd.timeout, key)
            except TimeoutError:
                logger.error("%s is not available." % device_tag)
                exit(1)

            if stdout:
                print(*stdout, sep='\n')
                sys.stdout.flush()

            if stderr:
                print(*stderr, sep='\n', file=sys.stderr)
                sys.stderr.flush()

            if not cmd.interactive:
                break

            if cmd_tokens and cmd_tokens[0] in EXIT_COMMANDS:
                break


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except (OSError, TimeoutError):
        logger.error("device is not available.")
        exit(1)

    except (EOFError, KeyboardInterrupt):
        print(file=sys.stderr)

    finally:
        StdIO.save_history(history_filename)

        if mqtt_client:
            mqtt_client.disconnect()

        if not cmd.interactive:
            exit(return_code)



"""
Created on 7 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse

from scs_core.aws.manager.configuration_finder import ConfigurationRequest


# --------------------------------------------------------------------------------------------------------------------

class CmdConfigurationCSV(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -n | -s | -l OUTPUT_CSV | "
                                                    "{ -d | -f } [-o OUTPUT_CSV_DIR] } [-t DEVICE_TAG [-e]] "
                                                    "[-v] [NODE_1..NODE_N]", version="%prog 1.0")

        # help...
        self.__parser.add_option("--node-names", "-n", action="store_true", dest="node_names", default=False,
                                 help="list the available nodes")

        # mode...
        self.__parser.add_option("--separate", "-s", action="store_true", dest="separate", default=False,
                                 help="retrieve latest configurations to separate CSVs (ignores NODEs)")

        self.__parser.add_option("--latest", "-l", type="string", action="store", dest="latest",
                                 help="retrieve latest configurations to single CSV")

        self.__parser.add_option("--diff-histories", "-d", action="store_true", dest="diff_histories", default=False,
                                 help="retrieve configuration history differences")

        self.__parser.add_option("--full-histories", "-f", action="store_true", dest="full_histories", default=False,
                                 help="retrieve full configuration histories")

        # filter...
        self.__parser.add_option("--device-tag", "-t", type="string", action="store", dest="device_tag",
                                 help="the device for the history report")

        self.__parser.add_option("--exactly", "-e", action="store_true", dest="exact_match", default=False,
                                 help="exact match for tag")

        # output...
        self.__parser.add_option("--output-csv-dir", "-o", type="string", action="store", dest="output_csv_dir",
                                 help="the directory in which to write histories")

        # narrative...
        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.node_names:
            count += 1

        if self.separate:
            count += 1

        if self.latest:
            count += 1

        if self.diff_histories:
            count += 1

        if self.full_histories:
            count += 1

        if count != 1:
            return False

        if self.device_tag is None and self.exact_match:
            return False

        return True


    def request_mode(self):
        if self.separate or self.latest:
            return ConfigurationRequest.Mode.LATEST

        if self.diff_histories:
            return ConfigurationRequest.Mode.DIFF

        if self.full_histories:
            return ConfigurationRequest.Mode.HISTORY

        return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def node_names(self):
        return self.__opts.node_names


    @property
    def separate(self):
        return self.__opts.separate


    @property
    def latest(self):
        return self.__opts.latest


    @property
    def diff_histories(self):
        return self.__opts.diff_histories


    @property
    def full_histories(self):
        return self.__opts.full_histories


    @property
    def device_tag(self):
        return self.__opts.device_tag


    @property
    def exact_match(self):
        return self.__opts.exact_match


    @property
    def output_csv_dir(self):
        return self.__opts.output_csv_dir


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def nodes(self):
        return self.__args


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdConfigurationCSV:{node_names:%s, separate:%s, latest:%s, diff_histories:%s, full_histories:%s, " \
               "device_tag:%s, exact_match:%s, output_csv_dir:%s, verbose:%s, nodes:%s}" %  \
               (self.node_names, self.separate, self.latest, self.diff_histories, self.full_histories,
                self.device_tag, self.exact_match, self.output_csv_dir, self.verbose, self.nodes)



#!/usr/bin/env python3

"""
Created on 10 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The organisations utility is used to

SYNOPSIS
organisations.py  { -F | -C -l LABEL -n LONG_NAME -u URL -o OWNER_EMAIL | \
-U LABEL [-l LABEL] [-n LONG_NAME] [-u URL] [-o OWNER_EMAIL] | -D LABEL } [-i INDENT] [-v]

EXAMPLES
organisations.py -F

DOCUMENT EXAMPLE
{"OrgID": 1, "Label": "SCS", "LongName": "South Coast Science", "URL": "https://www.southcoastscience.com",
"Owner": "bruno.beloff@southcoastscience.com"}

SEE ALSO
scs_analysis/cognito_credentials
"""

import requests
import sys

from scs_analysis.cmd.cmd_organisations import CmdOrganisations

from scs_core.aws.security.cognito_login_manager import CognitoUserLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials

from scs_core.aws.security.organisation import Organisation
from scs_core.aws.security.organisation_manager import OrganisationManager

from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.comms.stdio import StdIO
from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    credentials = None
    auth = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdOrganisations()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('organisations', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)

        if cmd.label is not None and not Organisation.is_valid_label(cmd.label):
            logger.error("the label '%s' is not valid." % cmd.label)
            exit(2)

        if cmd.long_name is not None and not Organisation.is_valid_long_name(cmd.long_name):
            logger.error("the long name '%s' is not valid." % cmd.long_name)
            exit(2)

        if cmd.url is not None and not Organisation.is_valid_url(cmd.url):
            logger.error("the URL '%s' is not valid." % cmd.url)
            exit(2)

        if cmd.owner is not None and not Organisation.is_valid_owner(cmd.owner):
            logger.error("the owner email address '%s' is not valid." % cmd.owner)
            exit(2)


        # ------------------------------------------------------------------------------------------------------------
        # auth...

        gatekeeper = CognitoUserLoginManager(requests)

        # CognitoUserCredentials...
        if not CognitoUserCredentials.exists(Host):
            logger.error("Cognito credentials not available.")
            exit(1)

        try:
            password = CognitoUserCredentials.password_from_user()
            credentials = CognitoUserCredentials.load(Host, encryption_key=password)
        except (KeyError, ValueError):
            logger.error("incorrect password")
            exit(1)

        try:
            auth = gatekeeper.login(credentials)

        except HTTPException as ex:
            logger.error(ex.data)
            exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        manager = OrganisationManager(requests)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.find:
            report = manager.find_organisations(auth.id_token)

        if cmd.create:
            org = Organisation(0, cmd.label, cmd.long_name, cmd.url, cmd.owner)
            report = manager.insert_organisation(auth.id_token, org)

        if cmd.update:
            # find...
            org = manager.get_organisation_by_label(auth.id_token, cmd.update)

            if org is None:
                logger.error("no organisation found for label: '%s'." % cmd.update)
                exit(1)

            # update...
            label = org.label if cmd.label is None else cmd.label
            long_name = org.long_name if cmd.long_name is None else cmd.long_name
            url = org.url if cmd.url is None else cmd.url
            owner = org.owner if cmd.owner is None else cmd.owner

            report = Organisation(org.org_id, label, long_name, url, owner)
            manager.update_organisation(auth.id_token, report)

        if cmd.delete:
            # find...
            org = manager.get_organisation_by_label(auth.id_token, cmd.delete)

            if org is None:
                logger.error("no organisation found for label: '%s'" % cmd.delete)
                exit(1)

            # check...
            response = StdIO.prompt('Are you sure (Y/n)?')
            if response != 'Y':
                exit(0)

            # delete...
            manager.delete_organisation(auth.id_token, cmd.delete)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        logger.error(ex.data)
        exit(1)



#!/usr/bin/env python3

"""
Created on 20 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The localised_datetime utility is used to report the localised date and time, as understood by the host operating
system, in ISO 8601 format. Optional offsets in hours, minutes and seconds can be supplied.

The utility is useful when composing topic history data requests, or just finding out what the time is somewhere else.

SYNOPSIS
localised_datetime.py { [-o HOURS] [-m MINUTES] [-s SECONDS] [-t TIMEZONE_NAME] | -z }

EXAMPLES
localised_datetime.py -t Europe/Athens
localised_datetime.py -o 2

DOCUMENT EXAMPLE - OUTPUT
2018-04-05T18:42:03.702+00:00

RESOURCES
https://en.wikipedia.org/wiki/ISO_8601
"""

import sys

from scs_analysis.cmd.cmd_localised_datetime import CmdLocalizedDatetime

from scs_core.data.datetime import LocalizedDatetime

from scs_core.location.timezone import Timezone


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdLocalizedDatetime()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("localised_datetime: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()


    # ----------------------------------------------------------------------------------------------------------------
    # run...

    if cmd.list:
        for zone in Timezone.zones():
            print(zone, file=sys.stderr)
        exit(0)

    now = LocalizedDatetime.now()
    offset = now.timedelta(hours=cmd.hours, minutes=cmd.minutes, seconds=cmd.seconds)

    if cmd.zone is not None:
        if not Timezone.is_valid(cmd.zone):
            print("localised_datetime: invalid timezone: %s" % cmd.zone, file=sys.stderr)
            exit(1)

        timezone = Timezone(cmd.zone)
        offset = offset.localize(timezone.zone)

    print(offset.as_iso8601())



"""
Created on 4 Sep 2020
Updated 23 Mar 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://packaging.python.org/tutorials/packaging-projects/
https://packaging.python.org/guides/single-sourcing-package-version/
"""

import codecs
import os
import setuptools


# --------------------------------------------------------------------------------------------------------------------

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            return line.split("'")[1]
    else:
        raise RuntimeError("Unable to find version string.")


# --------------------------------------------------------------------------------------------------------------------

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as req_txt:
    required = [line for line in req_txt.read().splitlines() if line]


setuptools.setup(
    name="scs-analysis",
    version=get_version("src/scs_analysis/__init__.py"),
    author="South Coast Science",
    author_email="contact@southcoastscience.com",
    description="Information management and analysis utilities for South Coast Science data consumers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/south-coast-science/scs_analysis",
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX",
    ],
    scripts=[
        'src/scs_analysis/access_key.py',
        'src/scs_analysis/alert.py',
        'src/scs_analysis/alert_status.py',
        'src/scs_analysis/aws_api_auth.py',
        'src/scs_analysis/aws_byline.py',
        'src/scs_analysis/aws_client_auth.py',
        'src/scs_analysis/aws_mqtt_client.py',
        'src/scs_analysis/aws_mqtt_control.py',
        'src/scs_analysis/aws_topic_history.py',
        'src/scs_analysis/aws_topic_publisher.py',
        'src/scs_analysis/aws_upload_interval.py',
        'src/scs_analysis/baseline.py',
        'src/scs_analysis/baseline_conf.py',
        'src/scs_analysis/configuration_csv.py',
        'src/scs_analysis/configuration_monitor.py',
        'src/scs_analysis/configuration_monitor_check.py',
        'src/scs_analysis/csv_collation_summary.py',
        'src/scs_analysis/csv_collator.py',
        'src/scs_analysis/csv_join.py',
        'src/scs_analysis/csv_reader.py',
        'src/scs_analysis/csv_segmentor.py',
        'src/scs_analysis/csv_writer.py',
        'src/scs_analysis/histo_chart.py',
        'src/scs_analysis/localised_datetime.py',
        'src/scs_analysis/monitor_auth.py',
        'src/scs_analysis/mqtt_peers.py',
        'src/scs_analysis/multi_chart.py',
        'src/scs_analysis/node.py',
        'src/scs_analysis/osio_api_auth.py',
        'src/scs_analysis/osio_client_auth.py',
        'src/scs_analysis/osio_mqtt_client.py',
        'src/scs_analysis/osio_mqtt_control.py',
        'src/scs_analysis/osio_topic_history.py',
        'src/scs_analysis/osio_topic_publisher.py',
        'src/scs_analysis/sample_aggregate.py',
        'src/scs_analysis/sample_average.py',
        'src/scs_analysis/sample_collator.py',
        'src/scs_analysis/sample_concentration.py',
        'src/scs_analysis/sample_distance.py',
        'src/scs_analysis/sample_duplicates.py',
        'src/scs_analysis/sample_error.py',
        'src/scs_analysis/sample_interval.py',
        'src/scs_analysis/sample_iso_8601.py',
        'src/scs_analysis/sample_low_pass.py',
        'src/scs_analysis/sample_max.py',
        'src/scs_analysis/sample_median.py',
        'src/scs_analysis/sample_midpoint.py',
        'src/scs_analysis/sample_min.py',
        'src/scs_analysis/sample_noise.py',
        'src/scs_analysis/sample_nullify.py',
        'src/scs_analysis/sample_paths.py',
        'src/scs_analysis/sample_regression.py',
        'src/scs_analysis/sample_slope.py',
        'src/scs_analysis/sample_sort.py',
        'src/scs_analysis/sample_stats.py',
        'src/scs_analysis/sample_subset.py',
        'src/scs_analysis/sample_time_shift.py',
        'src/scs_analysis/sample_timezone.py',
        'src/scs_analysis/single_chart.py',
        'src/scs_analysis/socket_receiver.py',
        'src/scs_analysis/timer.py',
        'src/scs_analysis/uds_receiver.py'
    ],
    install_requires=required,
    platforms=['any'],
    python_requires='>3.6',
    extras_require={
        'dev': [
            'pypandoc'
        ]
    }
)



"""
Created on 29 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse

from scs_core.data.recurring_period import RecurringPeriod


# --------------------------------------------------------------------------------------------------------------------

class CmdAlert(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog  { -F | -R ID | -C | -U ID | -D ID } "
                                                    "[-p TOPIC] [-f FIELD] [-l LOWER] [-u UPPER] "
                                                    "[-n { 1 | 0 }] [-a INTERVAL UNITS] [-t INTERVAL] [-s { 1 | 0 }] "
                                                    "[-c EMAIL_ADDR] [-e EMAIL_ADDR] [-i INDENT] [-v]",
                                              version="%prog 1.0")

        # operations...
        self.__parser.add_option("--find", "-F", action="store_true", dest="find", default=False,
                                 help="find alerts for given TOPIC (and FIELD) or creator email")

        self.__parser.add_option("--retrieve", "-R", type="int", action="store", dest="retrieve_id",
                                 help="retrieve alert with given ID")

        self.__parser.add_option("--create", "-C", action="store_true", dest="create", default=False,
                                 help="create alert")

        self.__parser.add_option("--update", "-U", type="int", action="store", dest="update_id",
                                 help="update alert with given ID")

        self.__parser.add_option("--delete", "-D", type="int", action="store", dest="delete_id",
                                 help="delete alert with given ID")

        # fields...
        self.__parser.add_option("--topic-path", "-p", type="string", action="store", dest="topic",
                                 help="topic path")

        self.__parser.add_option("--field", "-f", type="string", action="store", dest="field",
                                 help="field")

        self.__parser.add_option("--lower-threshold", "-l", type="float", action="store", dest="lower_threshold",
                                 help="lower threshold")

        self.__parser.add_option("--upper-threshold", "-u", type="float", action="store", dest="upper_threshold",
                                 help="upper threshold")

        self.__parser.add_option("--alert-on-none", "-n", type="int", action="store", dest="alert_on_none",
                                 default=False, help="alert on none (default false)")

        self.__parser.add_option("--aggregation-period", "-a", type="string", nargs=2, action="store",
                                 dest="aggregation_period", help="aggregation interval and units { D | H | M }")

        self.__parser.add_option("--test-interval", "-t", type="string", action="store", dest="test_interval",
                                 help="test interval")

        self.__parser.add_option("--suspended", "-s", type="int", action="store", dest="suspended",
                                 default=False, help="suspended (default false)")

        # email...
        self.__parser.add_option("--creator", "-c", type="string", action="store", dest="creator",
                                 help="email address of alert creator (admin use only)")

        self.__parser.add_option("--email-to", "-e", type="string", action="store", dest="to",
                                 help="email To address")

        # output...
        self.__parser.add_option("--indent", "-i", type="int", nargs=1, action="store", dest="indent",
                                 help="pretty-print the output with INDENT")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.find:
            count += 1

        if self.retrieve:
            count += 1

        if self.create:
            count += 1

        if self.update:
            count += 1

        if self.delete:
            count += 1

        if count != 1:
            return False

        if self.__opts.aggregation_period is not None:
            try:
                int(self.__opts.aggregation_period[0])
            except ValueError:
                return False

        if self.alert_on_none is not None and self.alert_on_none != 0 and self.alert_on_none != 1:
            return False

        if self.suspended is not None and self.suspended != 0 and self.suspended != 1:
            return False

        return True


    def is_complete(self):
        if not self.create:
            return True

        if self.topic is None or self.field is None or self.aggregation_period is None:
            return False

        if self.lower_threshold is None and self.upper_threshold is None and not self.alert_on_none:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def find(self):
        return self.__opts.find


    @property
    def retrieve_id(self):
        return self.__opts.retrieve_id


    @property
    def retrieve(self):
        return self.__opts.retrieve_id is not None


    @property
    def create(self):
        return self.__opts.create


    @property
    def update_id(self):
        return self.__opts.update_id


    @property
    def update(self):
        return self.__opts.update_id is not None


    @property
    def delete_id(self):
        return self.__opts.delete_id


    @property
    def delete(self):
        return self.__opts.delete_id is not None


    @property
    def topic(self):
        return self.__opts.topic


    @property
    def field(self):
        return self.__opts.field


    @property
    def lower_threshold(self):
        return self.__opts.lower_threshold


    @property
    def upper_threshold(self):
        return self.__opts.upper_threshold


    @property
    def alert_on_none(self):
        return self.__opts.alert_on_none


    @property
    def aggregation_period(self):
        period = self.__opts.aggregation_period
        return None if period is None else RecurringPeriod.construct(period[0], period[1])


    @property
    def test_interval(self):
        return self.__opts.test_interval


    @property
    def suspended(self):
        return self.__opts.suspended


    @property
    def creator(self):
        return self.__opts.creator


    @property
    def to(self):
        return self.__opts.to


    @property
    def indent(self):
        return self.__opts.indent


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdAlert:{find:%s, retrieve:%s, create:%s, update:%s, delete:%s, topic:%s, field:%s, " \
               "lower_threshold:%s, upper_threshold:%s, alert_on_none:%s, aggregation_period:%s, " \
               "test_interval:%s, suspended:%s, creator:%s, to:%s, indent:%s, verbose:%s}" % \
               (self.find, self.retrieve_id, self.create, self.update_id, self.delete_id, self.topic, self.field,
                self.lower_threshold, self.upper_threshold, self.alert_on_none, self.aggregation_period,
                self.test_interval, self.suspended, self.creator, self.to, self.indent, self.verbose)



#!/usr/bin/env python3

"""
Created on 4 Mar 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_regression utility provides a statical report on a single specified field for documents provided in the
input data stream. The specified field(s) normally represent an error or difference between two sources.

If the specified field(s) are not present in any of the input documents, the sample_regression utility terminates.
If a field is present but cannot be interpreted as a float, that document is ignored.

The output is a single JSON document. In standard mode, report fields are:

* Document count
* Minimum
* Mean
* Median
* Maximum
* Variance
* First standard deviation
* Second standard deviation
* Third standard deviation

In analytic mode, report fields are:

* Minimum
* Mean
* Median
* Maximum
* Lower boundary of 1st standard deviation
* Lower boundary of 2nd standard deviation
* Lower boundary of 3rd standard deviation
* Upper boundary of 1st standard deviation
* Upper boundary of 2nd standard deviation
* Upper boundary of 3rd standard deviation
* Amplitude by 1st standard deviation
* Amplitude by 2nd standard deviation
* Amplitude by 3rd standard deviation

A minimum of two input documents are required.

SYNOPSIS
sample_stats.py [-t] [-p PRECISION] [-a] [-r] [-v] PATH1 [PATH2 .. PATHN]

EXAMPLES
csv_reader.py -v scs-bgx-621-gases-2022-07-04-1min.csv | \
sample_stats.py -t -p1 -a val.VOC.cnc val.CO2.cnc exg.val.NO2.cnc | \
csv_writer.py -e scs-bgx-621-gases-2022-07-04-1min-stats.csv

DOCUMENT EXAMPLE - INPUT
{"ref": {"gas": {"SO2": 4.1}}, "SO2": {"vE": {"Urban": {"22Q1": 1.9}}},
"err": {"SO2": {"vE": {"Urban": {"22Q1": -2.2}}}}}
...

DOCUMENT EXAMPLE - OUTPUT
standard mode:
{"tag": "scs-bgx-619", "val": {"VOC": {"cnc": {"count": 4320, "min": 397.7, "mean": 448.2, "median": 446.1,
"max": 498.9, "var": 375.6, "stdev": 19.4, "stdev2": 38.8, "stdev3": 58.2}}}}

analytic mode:
{"tag": "scs-bgx-570", "val": {"SO2": {"cnc": {"min": -3.6, "mean": 1.7, "median": 1.4, "max": 9.3,
"l3": -5.8, "l2": -3.4, "l1": -1.0, "u1": 3.8, "u2": 6.2, "u3": 8.6, "a1": 4.8, "a2": 9.6, "a3": 14.4}

rows mode:
{"path": "val.NO2.cnc", "count": 60, "min": 13.1, "mean": 21.0, "median": 20.3, "max": 35.4, "var": 16.3,
"stdev": 4.0, "stdev2": 8.0, "stdev3": 12.0}
{"path": "val.Ox.cnc", "count": 60, "min": 98.9, "mean": 110.6, "median": 111.2, "max": 117.5, "var": 17.5,
"stdev": 4.2, "stdev2": 8.4, "stdev3": 12.6}
{"path": "val.NO.cnc", "count": 60, "min": 18.1, "mean": 28.1, "median": 25.4, "max": 69.7, "var": 92.0,
"stdev": 9.6, "stdev2": 19.2, "stdev3": 28.8}

RESOURCES
https://en.wikipedia.org/wiki/Standard_deviation
"""

import sys

from scs_analysis.cmd.cmd_sample_stats import CmdSampleStats

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict
from scs_core.data.stats import Stats, StatsAnalysis

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    tag = None
    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleStats()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('sample_stats', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        # data...
        values = {path: [] for path in cmd.paths}

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            if cmd.include_tag and tag is None:
                try:
                    tag = datum.node(cmd.tag)
                except KeyError:
                    tag = None

            document_count += 1

            for path in cmd.paths:
                if not datum.has_path(path):
                    logger.error("path '%s' not in datum: %s" % (path, line.strip()))
                    exit(1)

                try:
                    values[path].append(float(datum.node(path)))
                except ValueError:
                    continue

            processed_count += 1

        # validation...
        for path in cmd.paths:
            if len(values[path]) < 2:
                logger.error("at least two valid input documents are required for '%s'." % path)
                exit(1)

        # stats...
        report = PathDict()

        for path in cmd.paths:
            stats = Stats.construct(values[path], prec=cmd.precision)

            if cmd.analytic:
                stats = StatsAnalysis.construct_from_stats(stats, prec=cmd.precision)

            logger.info("%s: %s" % (path, stats))

            # output...
            stats_dict = PathDict(stats.as_json())

            if cmd.include_tag:
                report.append('tag', tag)

            if cmd.rows:
                report.append('path', path)
                for stats_path in stats_dict.paths():
                    report.append(stats_path, stats_dict.node(stats_path))

                print(JSONify.dumps(report))

            else:
                for stats_path in stats_dict.paths():
                    report.append('.'.join((path, stats_path)), stats_dict.node(stats_path))

        if not cmd.rows:
            print(JSONify.dumps(report))


# ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        logger.info("documents: %d processed: %d" % (document_count, processed_count))
